; Project name	:	FixYMF
; Description	:	Fixes Yamaha YMF71x hardware bug that requires
;					SB Pro Master Volume to be set to 1 so the other mixer
;					controls work.
; Copyright (C) 2022 by Tomi Tilli
;
; Uses assembly libraries from XTIDE Universal BIOS and Associated Tools
; Copyright (C) 2009-2010 by Tomi Tilli, 2011-2021 by XTIDE Universal BIOS Team.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; Visit http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
;

; Include .inc files
%define INCLUDE_STRING_LIBRARY
%define INCLUDE_DISPLAY_LIBRARY
%include "AssemblyLibrary.inc"	; Assembly Library. Must be included first!

SOUND_BLASTER_DEFAULT_BASE_PORT	EQU		220h
YMF_CTRL_DEFAULT_BASE_PORT		EQU		370h
IDX_YMF_CTRL_MASTER_VOL_LEFT	EQU		7h
IDX_YMF_CTRL_MASTER_VOL_RIGHT	EQU		8h
YMF_VALID_SB_MASTER_VOLUME		EQU		22h

SB_MIXER_DATA_REGISTER			EQU		5		; base + 5
IDX_CT1345_MASTER_VOLUME		EQU		22h		; Master Volume
CT1345_VOL_LEFT_SHIFT			EQU		5		; Bits 5, 6 and 7
CT1345_VOL_RIGHT_SHIFT			EQU		1		; Bits 1, 2 and 3


; Section containing code
SECTION .text

; Program first instruction.
ORG	100h						; Code starts at offset 100h (DOS .COM)
Start:
	jmp		FixYMF

; Include library and other sources
%include "AssemblyLibrary.asm"	; XTIDE Universal BIOS assembly library
%include "Print.asm"			; Functions to print strings (using Assembly library)


; Our variables parsed from SET BLASTER string
struc YMFVARS
	.wSBport				resb	2		; Sound Blaster base port
	.wCtrlPort				resb	2		; YMF control port
endstruc

ALIGN WORD_ALIGN
g_ymfVars:					; Our global variables
istruc YMFVARS
	at	YMFVARS.wSBport,	dw	SOUND_BLASTER_DEFAULT_BASE_PORT
	at	YMFVARS.wCtrlPort,	dw	YMF_CTRL_DEFAULT_BASE_PORT
iend
g_szDashForZero:			db	"- ",NULL		; Required by Assembly Library
g_szResultString:			db	"Mixer fixed from YMF71x at ports %x and %x",CR,LF,NULL
g_szErrorFailedToParse:		db	"Error! Failed to parse SET BLASTER string. Fix not applied!",CR,LF,NULL


;--------------------------------------------------------------------
; Program start
;--------------------------------------------------------------------
ALIGN JUMP_ALIGN
FixYMF:
	; Initialize libraries for printing
	cld
	CALL_DISPLAY_LIBRARY	InitializeDisplayContext
	call	Print_SetCharacterOutputToSTDOUT

	; Parse SET BLASTER string
	call	ParseSetBlasterStringFromPSP
	jc		SHORT .FailedToParseSetBlaster
	mov		[g_ymfVars+YMFVARS.wSBport], dx
	cmp		bp, -1
	je		SHORT .UseDefaultControlPort
	mov		[g_ymfVars+YMFVARS.wCtrlPort], bp
.UseDefaultControlPort:

	; Read SB Pro master volume to AL
	add		dx, SB_MIXER_DATA_REGISTER
	mov		ah, IDX_CT1345_MASTER_VOLUME
	call	SB_ReadToALfromMixerRegAH

	; Convert and write it to YMF control port
	call	SetYMF71xMasterVolume

	; Set SB master volume to 1
	mov		al, YMF_VALID_SB_MASTER_VOLUME
	call	SB_WriteALtoMixerRegAH
	sub		dx, SB_MIXER_DATA_REGISTER	; DX back to base port

	; All done. Print results
	mov		si, g_szResultString
	xchg	ax, dx
	mov		dx, [g_ymfVars+YMFVARS.wCtrlPort]
	call	Print_FormatStringFromSIwithParametersInAXDX
	xor		ax, ax				; AL = 0 = success
.ExitToDOSwithErrorCodeInAL:	; Without staying resident
	mov 	ah, 4Ch
	int 	21h					; Exit to DOS

.FailedToParseSetBlaster:
	mov		si, g_szErrorFailedToParse
	call	Print_NullTerminatedStringFromSI
	mov		al, 1
	jmp		SHORT .ExitToDOSwithErrorCodeInAL


;--------------------------------------------------------------------
; Parses BLASTER environment variable (SET BLASTER in autoexec.bat)
;
; ParseSetBlasterStringFromPSP
;	Parameters:
;		DS:		DOS PSP segment
;	Returns:
;		CL:		Bits 0...3 = Low DMA, bits 4...7 = High DMA (Fh if not used)
;		CH:		Bits 0...3 = IRQ, bits 4...7 = Sound Blaster type
;		DX:		Sound Blaster base port (FFFFh if not detected)
;		BP:		Control port (FFFFh if not detected)
;		CF:		Clear SET BLASTER parsed successfully
;				Set if error (invalid, incomplete or not found)
;	Corrupts registers:
;		AX, BX, SI, DI
;--------------------------------------------------------------------
ALIGN JUMP_ALIGN
ParseSetBlasterStringFromPSP:
	push	ds

	mov		ds, [2Ch]		; Offset 2Ch from DOS PSP = Segment address of environment area
	xor		si, si
	
	; We are looking for "BLASTER=A2xx Ix Dx Hx Tx" and it can also have lots
	; of other parameters as well, some unofficial (Unisound specific).
	; We are interested in A, I, D, H, T and R (Unisound) values
	xor		dx, dx					; Set return values to FFFFh
	dec		dx
	mov		cx, dx
	mov		bp, dx

.ScanForBlaster:
	mov		ax, [si]
	and		ax, 0DFDFh		; Convert to upper case
	cmp		ax, 'BL'
	jne		SHORT .FindNextNullTerminatedString
	inc		si
	inc		si

	mov		ax, [si]
	and		ax, 0DFDFh		; Convert to upper case
	cmp		ax, 'AS'
	jne		SHORT .FindNextNullTerminatedString
	inc		si
	inc		si

	mov		ax, [si]
	and		ax, 0DFDFh		; Convert to upper case
	cmp		ax, 'TE'
	jne		SHORT .FindNextNullTerminatedString
	inc		si
	inc		si

	mov		ax, [si]
	and		ax, 0FFDFh		; Convert to upper case
	cmp		ax, 'R='
	je		SHORT .ReadBlasterParameters

.FindNextNullTerminatedString:
	lodsb
	test	al, al
	jnz		SHORT .FindNextNullTerminatedString
	cmp		[si], al				; Another zero to end environment variables?
	jne		SHORT .ScanForBlaster	; If not, start again with next environment string
.ReturnNoValidBlasterStringFound:
	stc
.ReturnBlasterVariables:
	pop		ds
	ret

.ReadBlasterParameters:
	inc		si
	inc		si						; Point to one past '='
.ScanNextBlasterChar:
	lodsb							; Should be 'A' but we accept any order
	cmp		al, 'a'
	jb		SHORT .DoNotConvertToUpper
	cmp		al, 'z'
	ja		SHORT .DoNotConvertToUpper
	and		al, 0DFh				; To upper case
.DoNotConvertToUpper:
	cmp		al, 'A'
	je		SHORT .ReadPort
	cmp		al, 'I'
	je		SHORT .ReadInterrupt
	cmp		al, 'D'
	je		SHORT .ReadLowDMA
	cmp		al, 'H'
	je		SHORT .ReadHighDMA
	cmp		al, 'T'
	je		SHORT .ReadType
	cmp		al, 'R'
	je		SHORT .ReadControlPort		; Unisound specific
	test	al, al
	jnz		SHORT .ScanNextBlasterChar	; Space or unwanted parameter character

	; Zero. Terminate scan and verify what we got
	mov		ax, dx
	dec		ax
	jo		SHORT .ReturnNoValidBlasterStringFound
	; We could verify other variables here but we are happy with port only for now
	clc
	jmp		SHORT .ReturnBlasterVariables


.ReadPort:
	mov		bx, 16			; Port number is in hexadecimal
	call	ConvertBlasterStringValueToAXwithBaseInBX
	xchg	dx, ax			; Port number is returned in EDX low
	jmp		SHORT .ScanNextBlasterChar

.ReadInterrupt:
	call	ConvertBlasterStringValueToAX
	and		al, 0Fh
	and		ch, 0F0h
	or		ch, al			; Return interrupt in CH bits 0...3
	jmp		SHORT .ScanNextBlasterChar

.ReadLowDMA:
	call	ConvertBlasterStringValueToAX
	and		al, 0Fh
	and		cl, 0F0h
	or		cl, al			; Return low DMA in CL bits 0...3
	jmp		SHORT .ScanNextBlasterChar

.ReadHighDMA:
	call	ConvertBlasterStringValueToAX
	eSHL_IM	ax, 4
	and		cl, 0Fh
	or		cl, al			; Return high DMA in CL bits 4...7
	jmp		SHORT .ScanNextBlasterChar

.ReadType:
	call	ConvertBlasterStringValueToAX
	eSHL_IM	ax, 4
	and		ch, 0Fh
	or		ch, al			; Return type in CH bits 4...7
	jmp		SHORT .ScanNextBlasterChar

.ReadControlPort:
	mov		bx, 16			; Port number is in hexadecimal
	call	ConvertBlasterStringValueToAXwithBaseInBX
	xchg	bp, ax			; Return control port in BP
	jmp		.ScanNextBlasterChar


;--------------------------------------------------------------------
; Converts string to WORD. Stops at illegal character so string does
; not necessarily need to be NULL terminated. Any non decimal or
; hexadecimal character can terminate string.
;
; ConvertBlasterStringValueToAX	(base 10)
; ConvertBlasterStringValueToAXwithBaseInBX
;	Parameters:
;		BX:		Numeric base for conversion (10 or 16)
;		DS:SI:	Ptr to start of string
;	Returns:
;		AX:		WORD converted from string
;		SI:		Updated based on valid characters
;				SI points to first illegal character
;	Corrupts registers:
;		BX for ConvertBlasterStringValueToAX
;--------------------------------------------------------------------	
ConvertBlasterStringValueToAX:
	mov		bx, 10
ConvertBlasterStringValueToAXwithBaseInBX:
	push	cx
	call	String_ConvertWordToAXfromStringInDSSIwithBaseInBX
	dec		cx			; Invalid character where parsing stopped
	add		si, cx		; Number of valid characters processed
	pop		cx
	ret


;--------------------------------------------------------------------
; Reads from or writes to specified Sound Blaster Mixer register.
;
; SB_WriteALtoMixerRegAH
; SB_ReadToALfromMixerRegAH
;		Same as SB_MixerIO but modifies CL
; SB_MixerIO
;	Parameters:
;		AH:		Mixer register
;		AL:		Data to output to mixer register (if output)
;		CL:		bit 2: Set if output, otherwise input
;		DX:		Sound Blaster mixer data port
;	Returns:
;		AL:		Data from mixer register (if input)
;	Corrupts registers:
;		Nothing
;--------------------------------------------------------------------
SB_WriteALtoMixerRegAH:
	mov		cl, 1<<2
	jmp		SHORT SB_MixerIO
SB_ReadToALfromMixerRegAH:
	xor		cl, cl
	; Fall to SB_MixerIO
SB_MixerIO:
	; First we need to select the correct mixer register
	dec		dx			; Mixer Address port
	xchg	al, ah
	out		dx, al
	inc		dx			; DX to Mixer Data port
	xchg	al, ah

	; Input or output?
.NormalInputOrOutput:
	test	cx, (1<<2)
	jz		SHORT .InputFromMixerRegister

	; Output new value
	IO_DELAY
	out		dx, al
	jmp		SHORT .Return

.InputFromMixerRegister:
	in		al, dx
.Return:
	ret


;--------------------------------------------------------------------
; Convers Sound Blaster Pro master volume to YMF71x Control Register
; Master volume and sets it.
;
; SetYMF71xMasterVolume
;	Parameters:
;		AL:		Sound Blaster Master volume
;		BP:		YMF control port
;	Returns:
;		Nothing
;	Corrupts registers:
;		BX, DI
;--------------------------------------------------------------------
ALIGN JUMP_ALIGN
SetYMF71xMasterVolume:
	push	dx
	push	ax

	; Convert SB master volume to YMF master volume with lookup table
	mov		di, ax
	mov		bl, al
	eSHR_IM	bl, CT1345_VOL_LEFT_SHIFT
	shr		di, CT1345_VOL_RIGHT_SHIFT
	and		bx, 7h
	and		di, 7h
	mov		dx, bp								; Control port to DX

	mov		di, [di+.rgbSBmasterToYMFmaster]	; DI = right volume for YMF
	mov		ah, [bx+.rgbSBmasterToYMFmaster]	; AH = left volume for YMF

	; Output left master volume
	mov		al, IDX_YMF_CTRL_MASTER_VOL_LEFT
	call	.OutputNewVolumeWithIDXinALandVolInAH

	xchg	ax, di		; Right volume to AL...
	xchg	ah, al		; ...and to AH
	mov		al, IDX_YMF_CTRL_MASTER_VOL_RIGHT
	call	.OutputNewVolumeWithIDXinALandVolInAH

	pop		ax
	pop		dx
	ret

ALIGN JUMP_ALIGN
.OutputNewVolumeWithIDXinALandVolInAH:
	out		dx, al
	inc		dx		; YMF control data port
	xchg	al, ah	; New volume to AL
	IO_DELAY
	out		dx, al
	dec		dx		; YMF control base/index port
	ret

.rgbSBmasterToYMFmaster:
					db	8Fh	; SB master 0 => YMF master mute bit set and minimum volume
					db	0Eh	; SB master 1 => YMF master E = -28dB
					db	0Bh	; SB master 2 => YMF master B = -22dB
					db	08h	; SB master 3 => YMF master 8 = -16dB
					db	05h	; SB master 4 => YMF master 5 = -10dB
					db	03h	; SB master 5 => YMF master 3 = -6dB
					db	01h	; SB master 6 => YMF master 1 = -2dB
					db	00h	; SB master 7 => YMF master 0 = 0 dB
					db	0	; One extra byte to prevent reading past lookup table
